## This is an illustration of simulation.

import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
from functools import reduce
from SSVD.functions import SSVD_layer
from SSVD.functions import SSVD

def clusterheatmap(us, ss, vs, label):
    """Plot the clustered heatmap"""
    X_sparse = ss*us@vs.T
    levels = np.unique(label)
    row_index = np.empty(0, dtype = 'int')
    
    for i in range(len(levels)):
        index, = np.where(label == levels[i])
        index_sort = index[np.argsort(us.reshape(-1)[index])]
        row_index = np.concatenate((row_index, index_sort))
        
    v_reshape = vs.reshape(-1)
    col_sort = np.argsort(np.abs(v_reshape))
    X_sparse_sort = X_sparse[:,col_sort]
    col_index = np.argsort(v_reshape[col_sort])
    
    trans = (row_index.reshape(-1,1),col_index.reshape(1,-1))
    plot = sns.heatmap(X_sparse_sort[trans], vmin=-1, vmax=1, cmap = 'bwr')
    return plot


u_tilde = [list(range(10, 2, -1)), list(np.repeat(2, 17)), list(np.repeat(0, 75))]
u_tilde = np.array(reduce(lambda x, y: x+y, u_tilde))
v_tilde = [[10, -10, 8, -8, 5, -5], list(np.repeat(3, 5)), list(np.repeat(-3, 5)), list(np.repeat(0, 34))]
v_tilde = np.array(reduce(lambda x, y: x+y, v_tilde))

u_true = u_tilde / np.linalg.norm(u_tilde)  # (100,)
v_true = v_tilde / np.linalg.norm(v_tilde)  # (50,)
s_true = 50
Xstar = s_true * np.outer(u_true, v_true)  # (100, 50)

label_true = np.concatenate((np.ones(11-3), np.ones(17)*2, np.ones(75)*3))

plt.figure()
clusterheatmap(u_true[:,None], s_true, v_true[:,None], label_true)
plt.title("True SSVD first layer plot")
plt.show()

X = Xstar + np.random.normal(0, 1, Xstar.shape)
lam_grid = np.linspace(0, 8, 81)
gamma1 = gamma2 = 2
n_iters, us, vs, ss, lambda_us, lambda_vs = SSVD(X, 1, lam_grid, gamma1, gamma2)

plt.figure()
clusterheatmap(us, ss, vs, label_true)
plt.title("SSVD first layer plot by the codes")
plt.show()



## This is the second-step optimation, based on SSVDversion2. 
## Broadcasting are applied when selecting optimal lambda from a given grid. 
## Function get_tilde(), get_BIC(), and select_lam() are no longer necessary. 

import numpy as np 
import scipy.linalg as la

def update_uv(u_old, v_old, X, gamma1, gamma2, lam_grid):
    """Update u and v once given the current u and v.
    Input: 
        u_old: (n, 1)
        v_old: (d, 1)
        X: (n, d)
        Y: (nd, 1)
        Z: (nd, 1)
        gamma1, gamma2: scalar, tuning parameter
        lam_grid: ndarray, a grid of lambdas to be selected 
    Return:
        u_new: (n, 1)
        v_new: (d, 1)
        lambda_u, lambda_v: scalar, the optimal lambda under the current u and v
    """
    n, d = X.shape
    nd = n*d
    S = len(lam_grid)

    # initialize
    v_new = np.zeros(v_old.shape)
    u_new = np.zeros(u_old.shape)
    lambda_u = lambda_v = 0

    ## Update v using current u

    # ols for v, use current u
    v_tilde_hat = X.T @ u_old  # (d, 1), fixed ols estimate
    SSE_v = np.sum((X - np.outer(u_old, v_tilde_hat))**2)  # scalar, SSE_v = (Y-Y_hat).T @ (Y-Y_hat)
    sigma2_hat_v = SSE_v / (nd-d)  # scalar, fixed ols estimate

    # select lambda_v
    w2 = np.abs(v_tilde_hat)**(-gamma2)  # (d, 1)
    dfs_v = np.sum(np.abs(v_tilde_hat) > lam_grid*w2/2, axis=0)  # (S,), df for each lambda
    part2 = v_tilde_hat - lam_grid*w2/2; part2[part2<0] = 0
    part1 = np.sign(v_tilde_hat)
    v_tilde_br = part1 * part2  # (d, S), each column is v_tilde under each lambda
    outer_prods = np.outer(u_old, v_tilde_br.T)  # (n, d*S), each chunk of size (n, d) is the outer product mat under each lambda
    outer_prods = np.array(np.hsplit(outer_prods, S))  # (S, n, d)
    SSEs_v = ((X-outer_prods)**2).sum(axis = (1,2))  # (S,), ||X-uv_tilde^T||_F^2=||Y-Y_hat||^2 for each lambda
    BICs_v = SSEs_v/(nd*sigma2_hat_v) + np.log(nd)/nd*dfs_v  # (S,), BIC for each lambda
    lambda_v = lam_grid[np.argmin(BICs_v)]

    # update v
    part1 = np.sign(v_tilde_hat)
    part2 = np.abs(v_tilde_hat)-lambda_v*w2/2; part2[part2<0] = 0
    if not np.all(part2 == 0):  # not full shrinkage at v
        v_tilde = np.multiply(part1, part2)  # (d, 1)
        v_new = v_tilde / np.linalg.norm(v_tilde)
        
        ## Update u using current v

        # ols for u, use current v
        u_tilde_hat = X @ v_new  # (n, 1), fixed ols estimate
        SSE_u = np.sum((X.T - np.outer(v_new, u_tilde_hat))**2)
        sigma2_hat_u = SSE_u / (nd-n)  # scalar, fixed ols estimate

        # select lambda_u
        w1 = np.abs(u_tilde_hat)**(-gamma1)  # (n, 1)
        dfs_u = np.sum(np.abs(u_tilde_hat) > lam_grid*w1/2, axis=0)  # (S,), df for each lambda
        part2 = u_tilde_hat - lam_grid*w1/2; part2[part2<0] = 0
        part1 = np.sign(u_tilde_hat)
        u_tilde_br = part1 * part2  # (n, S), each column is u_tilde under each lambda
        outer_prods = np.outer(u_tilde_br.T, v_new)  # (n*S, d), each chunk of size (n, d) is the outer product mat under each lambda
        outer_prods = np.array(np.vsplit(outer_prods, S))  # (S, n, d)
        SSEs_u = ((X-outer_prods)**2).sum(axis = (1,2))  # (S,), ||X-u_tildev^T||_F^2=||Z-Z_hat||^2 for each lambda
        BICs_u = SSEs_u/(nd*sigma2_hat_u) + np.log(nd)/nd*dfs_u  # (S,), BIC for each lambda
        lambda_u = lam_grid[np.argmin(BICs_u)]

        # update u
        part1 = np.sign(u_tilde_hat)
        part2 = np.abs(u_tilde_hat)-lambda_u*w1/2; part2[part2<0] = 0
        if not np.all(part2 == 0):  # not full shrinkage at u
            u_tilde = np.multiply(part1, part2)  # (n, 1)
            u_new = u_tilde / np.linalg.norm(u_tilde)

    return u_new, v_new, lambda_u, lambda_v


def SSVD_layer(X, lam_grid, gamma1, gamma2, max_iter=5000, tol=1e-6):
    """Get the sparse SVD layer given the data matrix X at a SVD layer and the tuning parameters grid.
    Input: 
        X: (n, d), can be the original data matrix or the residual matrix
        lam_grid: ndarray, a grid of lambdas to be selected 
        gamma1, gamma2: scalar, tuning parameter
        max_iter: integer, the maximum iteration times
        tol: float, tolerance to stop the iteration
    Return: n_iter, u, v, s, lambda_u, lambda_v
        n_iter: number of iterations
        u: (n, 1), the final u at convergence 
        v: (d, 1), the final v at convergence 
        s: scalar, the singular value at convergence
        lambda_u, lambda_v: the optimal tuning parameter at convergence
    """
    # SVD
    U, _, VT = la.svd(X)

    # initial value
    u_old = U[:,0][:,None]
    v_old = VT[0][:,None]

    for i in range(max_iter):
        u_new, v_new, lambda_u, lambda_v = update_uv(u_old, v_old, X, gamma1, gamma2, lam_grid)
        if np.linalg.norm(u_new-u_old) < tol and np.linalg.norm(v_new-v_old) < tol:  # achieve the tolerance
            break 
        if np.all(u_new == 0) or np.all(u_new == 0):  # full shrinkage (i.e., all zeros in the vector)
            print("Warning: Full shrinkage has been achieved. Iterations stops. No further decomposition. The desired number of layers may not be achieved. ")
            break
        u_old, v_old = u_new, v_new
    n_iter = i+1  # number of iterations
    u, v = u_new, v_new  # the final u and v at convergence 
    s = (u_new.T @ X @ v_new)[0][0]
    if n_iter == max_iter:
        print("Warning: The maximum iteration has been achieved. Please consider increasing `max_iter`.")
    return n_iter, u, v, s, lambda_u, lambda_v

def SSVD(X, num_layer, lam_grid, gamma1, gamma2, max_iter=5000, tol=1e-6):
    """Get the SSVD given the data matrix X and the desired number of SSVD layers.
    Input: 
        X: (n, d), the original data matrix
        num_layer: desired number of SSVD layers
        lam_grid: ndarray, a grid of lambdas to be selected 
        gamma1, gamma2: scalar, tuning parameter
        max_iter: integer, the maximum iteration times
        tol: float, tolerance to stop the iteration
    Return: n_iters, us, vs, ss, lambda_us, lambda_vs
        n_iters: (num_layer,), number of iterations for each layer
        us: (n, num_layer), the final u at convergence for each layer
        vs: (d, num_layer), the final v at convergence for each layer
        ss: (num_layer,), the singular value at convergence for each layer
        lambda_us, lambda_vs: (num_layer,), the optimal tuning parameter at convergence for each layer
    """
    n, d = X.shape
    n_iters = np.zeros(num_layer, dtype = int)
    ss = np.zeros(num_layer)
    lambda_us = np.zeros(num_layer)
    lambda_vs = np.zeros(num_layer)
    us = np.zeros((n, num_layer))
    vs = np.zeros((d, num_layer))
    # initial value
    u = np.zeros((n, 1)); v = np.zeros((d, 1)); resi_mat = X; s = 0
    for i in range(num_layer):
        resi_mat = resi_mat - s*u@v.T
        n_iter, u, v, s, lambda_u, lambda_v = SSVD_layer(resi_mat, lam_grid, gamma1, gamma2, max_iter, tol)
        n_iters[i] = n_iter
        ss[i] = s
        lambda_us[i] = lambda_u
        lambda_vs[i] = lambda_v 
        us[:,i] = u[:,0]
        vs[:,i] = v[:,0]
        if np.all(u == 0) or np.all(v == 0):  # full shrinkage (i.e., all zeros in the vector)
            break
    return n_iters, us, vs, ss, lambda_us, lambda_vs



## This is the final version optimized by Cython.
## The following should be run in Jupyter notebook. 

import cython
%load_ext cython

%%cython

import numpy as np 
import scipy.linalg as la
import cython
from cython.parallel import parallel, prange

cdef extern from "math.h":
    double log(double x) nogil
    double pow(double x, double y) nogil
    double fabs(double x) nogil
    double sqrt(double x)
    double isless(double x, double y) nogil
    double fmax(double x, double y)
    double fma(double x, double y, double z) nogil


cdef double vector_dist_sq(double[:,:] u, double[:,:] v):
    """Squared Euclidean distance between two vectors. Can also compute the squared norm of a vector if set one of the input being zero vector. """
    cdef int i 
    cdef double s = 0 
    for i in range(u.shape[0]):
        s += pow(u[i,0] - v[i,0], 2)
    return s

@cython.boundscheck(False)
@cython.wraparound(False)
cdef matrix_multiply(double[:,:] u, double[:, :] v, double[:, :] res, double c = 1):
    """Matrix multiplication, equivalent to c*u@v."""
    cdef int i, j, k
    cdef int m = u.shape[0], n = u.shape[1], p = v.shape[1]
    with cython.nogil, parallel():
        for i in prange(m):  # parallel
            for j in prange(p):  # parallel
                res[i,j] = 0
                for k in range(n):  # serial
                    res[i,j] += u[i,k] * v[k,j]
                res[i,j] = fma(res[i,j], c, 0)

@cython.boundscheck(False)
@cython.wraparound(False)
cdef elementwise_multiply(double[:,:] u, double[:,:] v, double[:,:] res):
    """Element multiplication of two matrices. """
    cdef int i, j
    cdef int m = u.shape[0], n = u.shape[1]
    with cython.nogil, parallel():
        for i in prange(m):  # parallel
            for j in prange(n):  # parallel
                res[i,j] = u[i,j] * v[i,j]


@cython.boundscheck(False)
@cython.wraparound(False)
cdef double[:,:] vec_outer_prod(double[:,:] u, double[:,:] v, double c = 1):
    """Outer product between two vectors, equivalent to c*u@v.T."""
    cdef int n = u.shape[0], m = v.shape[0]
    cdef int i, j
    cdef double[:,:] res = np.zeros((n, m))
    with cython.nogil, parallel():
        for i in prange(n):  # parallel
            for j in prange(m):  # parallel
                res[i,j] = u[i,0] * v[j,0] * c
    return res


@cython.boundscheck(False)
@cython.wraparound(False)
cdef double[:,:] get_w(double[:,:] tilde_hat, double gamma):
    """Get the w vector, equivalent to elementwise |tilde_hat|^(-gamma)"""
    cdef int i, l = tilde_hat.shape[0]
    cdef double[:,:] w = np.zeros((l, 1))
    with cython.nogil, parallel():
        for i in prange(l):  # parallel
            w[i,0] = pow(fabs(tilde_hat[i,0]), -gamma)
    return w



@cython.boundscheck(False)
@cython.wraparound(False)
cdef double[:,:] get_part2(double[:,:] tilde_hat, double[:,:] w, double lam):
    """Compute the part2 in the updating formula. """
    cdef int i, l = tilde_hat.shape[0]
    cdef double ele 
    cdef double[:,:] part2 = np.zeros((l,1))

    with cython.nogil, parallel():
        for i in prange(l):  # parallel
            ele = fabs(tilde_hat[i,0])-lam*w[i,0]/2
            if ele > 0:
                part2[i,0] = ele 
    return part2
cdef int is_full_shinkage(double[:,:] v):
    """Judge if the input vector if fully shrunk to 0. Return 1 if fully shunk, and 0 otherwise. """
    cdef int l = v.shape[0]
    cdef int flag = 1
    
    for i in range(l):
        if v[i,0] != 0:
            flag = 0
            break
    return flag  # flag = 1 if all elements are 0, = 0 otherwise

@cython.boundscheck(False)
@cython.wraparound(False)
cdef get_vec_new(double[:,:] tilde_vec, double[:,:] vec_new):
    """Get the vec_new according to tilde_vec/norm(tilde_vec). """
    cdef int i, l = tilde_vec.shape[0]
    cdef double norm = sqrt(vector_dist_sq(tilde_vec, np.zeros((l, 1))))
    with cython.nogil, parallel():
        for i in prange(l):  # parallel
            vec_new[i,0] = tilde_vec[i,0] / norm


cpdef update_uv3c(double[:,:] u_old, X, gamma1, gamma2, lam_grid, double[:,:] u_new, double[:,:] v_new):
    """Update u and v once given the current u and v."""
    cdef int n = X.shape[0], d = X.shape[1], S = lam_grid.shape[0]
    cdef int nd = n*d

    cdef double[:,:] v_tilde_hat = np.zeros((d,1)), v_tilde = np.zeros((d,1))
    cdef double[:,:] u_tilde_hat = np.zeros((n,1)), u_tilde = np.zeros((n,1))

    cdef double lambda_u = 0, lambda_v = 0


    ## Update v using current u

    # ols for v, use current u
    matrix_multiply(X.T, u_old, v_tilde_hat)  # v_tilde_hat: (d, 1), fixed ols estimate
    SSE_v = np.sum((X - vec_outer_prod(u_old, v_tilde_hat))**2)  # scalar, SSE_v = (Y-Y_hat).T @ (Y-Y_hat)
    sigma2_hat_v = SSE_v / (nd-d)  # scalar, fixed ols estimate

    # select lambda_v
    w2 = get_w(v_tilde_hat, gamma2)  # (d, 1)
    dfs_v = np.sum(np.abs(v_tilde_hat) > lam_grid*w2/2, axis=0)  # (S,), df for each lambda
    part2 = v_tilde_hat - lam_grid*w2/2; part2[part2<0] = 0
    part1 = np.sign(v_tilde_hat)
    v_tilde_br = part1 * part2  # (d, S), each column is v_tilde under each lambda
    outer_prods = np.outer(u_old, v_tilde_br.T)  # (n, d*S), each chunk of size (n, d) is the outer product mat under each lambda
    outer_prods = np.array(np.hsplit(outer_prods, S))  # (S, n, d)
    SSEs_v = ((X-outer_prods)**2).sum(axis = (1,2))  # (S,), ||X-uv_tilde^T||_F^2=||Y-Y_hat||^2 for each lambda
    BICs_v = SSEs_v/(nd*sigma2_hat_v) + np.log(nd)/nd*dfs_v  # (S,), BIC for each lambda
    lambda_v = lam_grid[np.argmin(BICs_v)]

    # update v
    part1 = np.sign(v_tilde_hat)
    part2 = get_part2(v_tilde_hat, w2, lambda_v)

    if is_full_shinkage(part2) == 0:  # not full shrinkage at v
        elementwise_multiply(part1, part2, v_tilde)  # v_tilde: (d, 1)
        get_vec_new(v_tilde, v_new)  # v_new: (d, 1)
        
        ## Update u using current v

        # ols for u, use current v
        matrix_multiply(X, v_new, u_tilde_hat)  # u_tilde_hat: (n, 1), fixed ols estimate
        SSE_u = np.sum((X.T - vec_outer_prod(v_new, u_tilde_hat))**2)
        sigma2_hat_u = SSE_u / (nd-n)  # scalar, fixed ols estimate

        # select lambda_u
        w1 = get_w(u_tilde_hat, gamma1)  # (n, 1)
        dfs_u = np.sum(np.abs(u_tilde_hat) > lam_grid*w1/2, axis=0)  # (S,), df for each lambda
        part2 = u_tilde_hat - lam_grid*w1/2; part2[part2<0] = 0
        part1 = np.sign(u_tilde_hat)
        u_tilde_br = part1 * part2  # (n, S), each column is u_tilde under each lambda
        outer_prods = np.outer(u_tilde_br.T, v_new)  # (n*S, d), each chunk of size (n, d) is the outer product mat under each lambda
        outer_prods = np.array(np.vsplit(outer_prods, S))  # (S, n, d)
        SSEs_u = ((X-outer_prods)**2).sum(axis = (1,2))  # (S,), ||X-u_tildev^T||_F^2=||Z-Z_hat||^2 for each lambda
        BICs_u = SSEs_u/(nd*sigma2_hat_u) + np.log(nd)/nd*dfs_u  # (S,), BIC for each lambda
        lambda_u = lam_grid[np.argmin(BICs_u)]

        # update u
        part1 = np.sign(u_tilde_hat)
        part2 = get_part2(u_tilde_hat, w1, lambda_u)
        if is_full_shinkage(part2) == 0:  # not full shrink at u
            elementwise_multiply(part1, part2, u_tilde)  # u_tilde: (n, 1)
            get_vec_new(u_tilde, u_new)  # u_new: (n, 1)

    return lambda_u, lambda_v


cpdef SSVD_layer3c(X, lam_grid, gamma1, gamma2, max_iter=5000, tol=1e-6):
    """Get the sparse SVD layer given the data matrix X at a SVD layer and the tuning parameters grid."""
    cdef int n = X.shape[0], d = X.shape[1]
    # SVD
    U, _, VT = la.svd(X)

    # initial value
    cdef double[:,:] u_old = U[:,0][:,None], v_old = VT[0][:,None]
    cdef double[:,:] u_new = np.zeros((n,1)), v_new = np.zeros((d,1))
    cdef int i  # number of iterations

    for i in range(max_iter):
        lambda_u, lambda_v = update_uv3c(u_old, X, gamma1, gamma2, lam_grid, u_new, v_new)  # update u_new, v_new
        if isless(fmax(vector_dist_sq(u_new, u_old), vector_dist_sq(v_new, v_old)), pow(tol, 2)):  # achieve the tolerance
            break 
        if fmax(is_full_shinkage(u_new), is_full_shinkage(v_new)):  # full shrinkage (i.e., all zeros in the vector)
            print("Warning: Full shrinkage has been achieved. Iterations stops. No further decomposition. The desired number of layers may not be achieved. ")
            break
        u_old, v_old = u_new, v_new
    u, v = u_new, v_new  # the final u and v at convergence 
    s = (u.T @ X @ v)[0][0]
    n_iter = i+1
    if n_iter == max_iter:
        print("Warning: The maximum iteration has been achieved. Please consider increasing `max_iter`.")
    return n_iter, np.array(u), np.array(v), s, lambda_u, lambda_v 


cpdef SSVD3c(X, num_layer, lam_grid, gamma1, gamma2, max_iter=5000, tol=1e-6):
    """Get the SSVD given the data matrix X and the desired number of SSVD layers."""
    n, d = X.shape
    n_iters = np.zeros(num_layer, dtype = int)
    ss = np.zeros(num_layer)
    lambda_us = np.zeros(num_layer)
    lambda_vs = np.zeros(num_layer)
    us = np.zeros((n, num_layer))
    vs = np.zeros((d, num_layer))
    # initial value
    cdef double[:,:] res = np.zeros((n, d))
    cdef double s = 0
    resi_mat = X
    u = np.zeros((n, 1)); v = np.zeros((d, 1))
    for i in range(num_layer):
        resi_mat = resi_mat - s * np.outer(u, v)
        n_iter, u, v, s, lambda_u, lambda_v = SSVD_layer3c(resi_mat, lam_grid, gamma1, gamma2, max_iter, tol)
        n_iters[i] = n_iter
        ss[i] = s
        lambda_us[i] = lambda_u
        lambda_vs[i] = lambda_v 
        us[:,i] = u[:,0]
        vs[:,i] = v[:,0]
        if np.all(u == 0) or np.all(v == 0):  # full shrinkage (i.e., all zeros in the vector)
            break
    return n_iters, us, vs, ss, lambda_us, lambda_vs


## This is the most efficient version, corresponding to version 3 in the report,
## before going into Cython.

import numpy as np 
import scipy.linalg as la 
import pandas as pd 
import matplotlib.pyplot as plt
import seaborn as sns
import numpy as np 
from functools import reduce
from sklearn.decomposition import SparsePCA


def clusterheatmap(us, ss, vs, label):
    """Plot the clustered heatmap"""
    X_sparse = ss*us@vs.T
    levels = np.unique(label)
    row_index = np.empty(0, dtype = 'int')
    
    for i in range(len(levels)):
        index, = np.where(label == levels[i])
        index_sort = index[np.argsort(us.reshape(-1)[index])]
        row_index = np.concatenate((row_index, index_sort))
        
    v_reshape = vs.reshape(-1)
    col_sort = np.argsort(np.abs(v_reshape))
    X_sparse_sort = X_sparse[:,col_sort]
    col_index = np.argsort(v_reshape[col_sort])
    
    trans = (row_index.reshape(-1,1),col_index.reshape(1,-1))
    plot = sns.heatmap(X_sparse_sort[trans], vmin=-1, vmax=1, cmap = 'bwr')
    return plot


def update_uv(u_old, v_old, X, gamma1, gamma2, lam_grid):
    """Update u and v once given the current u and v.
    Input: 
        u_old: (n, 1)
        v_old: (d, 1)
        X: (n, d)
        Y: (nd, 1)
        Z: (nd, 1)
        gamma1, gamma2: scalar, tuning parameter
        lam_grid: ndarray, a grid of lambdas to be selected 
    Return:
        u_new: (n, 1)
        v_new: (d, 1)
        lambda_u, lambda_v: scalar, the optimal lambda under the current u and v
    """
    n, d = X.shape
    nd = n*d
    S = len(lam_grid)

    # initialize
    v_new = np.zeros(v_old.shape)
    u_new = np.zeros(u_old.shape)
    lambda_u = lambda_v = 0

    ## Update v using current u

    # ols for v, use current u
    v_tilde_hat = X.T @ u_old  # (d, 1), fixed ols estimate
    SSE_v = np.sum((X - np.outer(u_old, v_tilde_hat))**2)  # scalar, SSE_v = (Y-Y_hat).T @ (Y-Y_hat)
    sigma2_hat_v = SSE_v / (nd-d)  # scalar, fixed ols estimate

    # select lambda_v
    w2 = np.abs(v_tilde_hat)**(-gamma2)  # (d, 1)
    dfs_v = np.sum(np.abs(v_tilde_hat) > lam_grid*w2/2, axis=0)  # (S,), df for each lambda
    part2 = v_tilde_hat - lam_grid*w2/2; part2[part2<0] = 0
    part1 = np.sign(v_tilde_hat)
    v_tilde_br = part1 * part2  # (d, S), each column is v_tilde under each lambda
    outer_prods = np.outer(u_old, v_tilde_br.T)  # (n, d*S), each chunk of size (n, d) is the outer product mat under each lambda
    outer_prods = np.array(np.hsplit(outer_prods, S))  # (S, n, d)
    SSEs_v = ((X-outer_prods)**2).sum(axis = (1,2))  # (S,), ||X-uv_tilde^T||_F^2=||Y-Y_hat||^2 for each lambda
    BICs_v = SSEs_v/(nd*sigma2_hat_v) + np.log(nd)/nd*dfs_v  # (S,), BIC for each lambda
    lambda_v = lam_grid[np.argmin(BICs_v)]

    # update v
    part1 = np.sign(v_tilde_hat)
    part2 = np.abs(v_tilde_hat)-lambda_v*w2/2; part2[part2<0] = 0
    if not np.all(part2 == 0):  # not full shrinkage at v
        v_tilde = np.multiply(part1, part2)  # (d, 1)
        v_new = v_tilde / np.linalg.norm(v_tilde)
        
        ## Update u using current v

        # ols for u, use current v
        u_tilde_hat = X @ v_new  # (n, 1), fixed ols estimate
        SSE_u = np.sum((X.T - np.outer(v_new, u_tilde_hat))**2)
        sigma2_hat_u = SSE_u / (nd-n)  # scalar, fixed ols estimate

        # select lambda_u
        w1 = np.abs(u_tilde_hat)**(-gamma1)  # (n, 1)
        dfs_u = np.sum(np.abs(u_tilde_hat) > lam_grid*w1/2, axis=0)  # (S,), df for each lambda
        part2 = u_tilde_hat - lam_grid*w1/2; part2[part2<0] = 0
        part1 = np.sign(u_tilde_hat)
        u_tilde_br = part1 * part2  # (n, S), each column is u_tilde under each lambda
        outer_prods = np.outer(u_tilde_br.T, v_new)  # (n*S, d), each chunk of size (n, d) is the outer product mat under each lambda
        outer_prods = np.array(np.vsplit(outer_prods, S))  # (S, n, d)
        SSEs_u = ((X-outer_prods)**2).sum(axis = (1,2))  # (S,), ||X-u_tildev^T||_F^2=||Z-Z_hat||^2 for each lambda
        BICs_u = SSEs_u/(nd*sigma2_hat_u) + np.log(nd)/nd*dfs_u  # (S,), BIC for each lambda
        lambda_u = lam_grid[np.argmin(BICs_u)]

        # update u
        part1 = np.sign(u_tilde_hat)
        part2 = np.abs(u_tilde_hat)-lambda_u*w1/2; part2[part2<0] = 0
        if not np.all(part2 == 0):  # not full shrinkage at u
            u_tilde = np.multiply(part1, part2)  # (n, 1)
            u_new = u_tilde / np.linalg.norm(u_tilde)

    return u_new, v_new, lambda_u, lambda_v


def SSVD_layer(X, lam_grid, gamma1, gamma2, max_iter=5000, tol=1e-6):
    """Get the sparse SVD layer given the data matrix X at a SVD layer and the tuning parameters grid.
    Input: 
        X: (n, d), can be the original data matrix or the residual matrix
        lam_grid: ndarray, a grid of lambdas to be selected 
        gamma1, gamma2: scalar, tuning parameter
        max_iter: integer, the maximum iteration times
        tol: float, tolerance to stop the iteration
    Return: n_iter, u, v, s, lambda_u, lambda_v
        n_iter: number of iterations
        u: (n, 1), the final u at convergence 
        v: (d, 1), the final v at convergence 
        s: scalar, the singular value at convergence
        lambda_u, lambda_v: the optimal tuning parameter at convergence
    """
    # SVD
    U, _, VT = la.svd(X)

    # initial value
    u_old = U[:,0][:,None]
    v_old = VT[0][:,None]

    for i in range(max_iter):
        u_new, v_new, lambda_u, lambda_v = update_uv(u_old, v_old, X, gamma1, gamma2, lam_grid)
        if np.linalg.norm(u_new-u_old) < tol and np.linalg.norm(v_new-v_old) < tol:  # achieve the tolerance
            break 
        if np.all(u_new == 0) or np.all(u_new == 0):  # full shrinkage (i.e., all zeros in the vector)
            print("Warning: Full shrinkage has been achieved. Iterations stops. No further decomposition. The desired number of layers may not be achieved. ")
            break
        u_old, v_old = u_new, v_new
    n_iter = i+1  # number of iterations
    u, v = u_new, v_new  # the final u and v at convergence 
    s = (u_new.T @ X @ v_new)[0][0]
    if n_iter == max_iter:
        print("Warning: The maximum iteration has been achieved. Please consider increasing `max_iter`.")
    return n_iter, u, v, s, lambda_u, lambda_v

def SSVD(X, num_layer, lam_grid, gamma1, gamma2, max_iter=5000, tol=1e-6):
    """Get the SSVD given the data matrix X and the desired number of SSVD layers.
    Input: 
        X: (n, d), the original data matrix
        num_layer: desired number of SSVD layers
        lam_grid: ndarray, a grid of lambdas to be selected 
        gamma1, gamma2: scalar, tuning parameter
        max_iter: integer, the maximum iteration times
        tol: float, tolerance to stop the iteration
    Return: n_iters, us, vs, ss, lambda_us, lambda_vs
        n_iters: (num_layer,), number of iterations for each layer
        us: (n, num_layer), the final u at convergence for each layer
        vs: (d, num_layer), the final v at convergence for each layer
        ss: (num_layer,), the singular value at convergence for each layer
        lambda_us, lambda_vs: (num_layer,), the optimal tuning parameter at convergence for each layer
    """
    n, d = X.shape
    n_iters = np.zeros(num_layer, dtype = int)
    ss = np.zeros(num_layer)
    lambda_us = np.zeros(num_layer)
    lambda_vs = np.zeros(num_layer)
    us = np.zeros((n, num_layer))
    vs = np.zeros((d, num_layer))
    # initial value
    u = np.zeros((n, 1)); v = np.zeros((d, 1)); resi_mat = X; s = 0
    for i in range(num_layer):
        resi_mat = resi_mat - s*u@v.T
        n_iter, u, v, s, lambda_u, lambda_v = SSVD_layer(resi_mat, lam_grid, gamma1, gamma2, max_iter, tol)
        n_iters[i] = n_iter
        ss[i] = s
        lambda_us[i] = lambda_u
        lambda_vs[i] = lambda_v 
        us[:,i] = u[:,0]
        vs[:,i] = v[:,0]
        if np.all(u == 0) or np.all(v == 0):  # full shrinkage (i.e., all zeros in the vector)
            break
    return n_iters, us, vs, ss, lambda_us, lambda_vs



## This the most basic realization of the SSVD algorithm.
## Codes directly correspond to the paper context. 

import numpy as np 
import scipy.linalg as la


def get_tilde(lam, tilde_hat, w):
    """Get the tilde vector given the lambda, tilde_hat_vector, and w."""
    part1 = np.sign(tilde_hat)
    part2 = np.abs(tilde_hat)-lam*w/2
    part2[part2<np.zeros(part2.shape)] = 0
    tilde = np.multiply(part1, part2)
    return tilde

def get_BIC(lam, w, tilde_hat, fixed_vec, response, sigma2_hat, nd):
    """ Compute BIC given the lambda.
    lam: scalar
    For v:
        w = w2: (d, 1)
        tilde_hat = v_tilde_hat: (d, 1)
        fixed_vec = u_old: (n, 1)
        response = Y: (nd, 1)
        sigma2_hat = sigma2_hat_v: scalar
        nd: scalar, =n*d
    For u:
        w = w1: (n, 1)
        tilde_hat = u_tilde_hat: (n, 1)
        fixed_vec = v_new: (d, 1)
        response = Z: (nd, 1)
        sigma2_hat = sigma2_hat_u: scalar
        nd: scalar, =n*d
    Return: BIC, scalar
    """
    df = np.sum(np.abs(tilde_hat) > lam*w/2)
    tilde = get_tilde(lam, tilde_hat, w)  # get the tilde vector udner the current lambda
    response_hat = np.kron(np.eye(tilde.shape[0]), fixed_vec) @ tilde
    BIC = np.linalg.norm(response-response_hat)**2/(nd*sigma2_hat)+np.log(nd)/nd*df
    return BIC
def select_lam(lam_grid, w, tilde_hat, fixed_vec, response, sigma2_hat, nd):
    """ Select lambda given a lambda grid based on BIC.
    lam_grid: (S,)
    For v:
        w = w2: (d, 1)
        tilde_hat = v_tilde_hat: (d, 1)
        fixed_vec = u_old: (n, 1)
        response = Y: (nd, 1)
        sigma2_hat = sigma2_hat_v: scalar
        nd: scalar, =n*d
    For u:
        w = w1: (n, 1)
        tilde_hat = u_tilde_hat: (n, 1)
        fixed_vec = v_new: (d, 1)
        response = Z: (nd, 1)
        sigma2_hat = sigma2_hat_u: scalar
        nd: scalar, =n*d
    Return: scalar, the optimal lambda
    """
    BICs = list(map(lambda lam: get_BIC(lam, w, tilde_hat, fixed_vec, response, sigma2_hat, nd), lam_grid))
    return lam_grid[np.argmin(BICs)]
def update_uv(u_old, v_old, X, Y, Z, gamma1, gamma2, lam_grid):
    """Update u and v once given the current u and v.
    Input: 
        u_old: (n, 1)
        v_old: (d, 1)
        X: (n, d)
        Y: (nd, 1)
        Z: (nd, 1)
        gamma1, gamma2: scalar, tuning parameter
        lam_grid: ndarray, a grid of lambdas to be selected 
    Return:
        u_new: (n, 1)
        v_new: (d, 1)
        lambda_u, lambda_v: scalar, the optimal lambda under the current u and v
    """
    n, d = X.shape
    nd = n*d

    ## Update v using current u

    # ols for v, use current u
    v_tilde_hat = X.T @ u_old  # (d, 1), fixed ols estimate
    Y_hat = (np.kron(np.eye(d), u_old) @ v_tilde_hat).reshape((-1,1))  # (nd, 1), ols estimate
    sigma2_hat_v = ((Y-Y_hat).T @ (Y-Y_hat) / (nd-d))[0][0]  # scalar, fixed ols estimate

    # select lambda_v
    w2 = np.abs(v_tilde_hat)**(-gamma2)  # (d, 1)
    lambda_v = select_lam(lam_grid, w2, v_tilde_hat, u_old, Y, sigma2_hat_v, nd)

    # update v
    v_tilde = get_tilde(lambda_v, v_tilde_hat, w2)  # (d, 1)

    if np.all(v_tilde == 0):  # full shrinkage at v
        v_new = np.zeros(v_old.shape)
        u_new = np.zeros(u_old.shape)
        lambda_u = 0
    else:
        v_new = v_tilde / np.linalg.norm(v_tilde)

        ## Update u using current v

        # ols for u, use current v
        u_tilde_hat = X @ v_new  # (n, 1), fixed ols estimate
        Z_hat = np.kron(np.eye(n), v_new) @ u_tilde_hat  # (nd, 1), ols estimate
        sigma2_hat_u = ((Z-Z_hat).T @ (Z-Z_hat) / (nd-n))[0][0]  # scalar, fixed ols estimate

        # select lambda_u
        w1 = np.abs(u_tilde_hat)**(-gamma1)
        lambda_u = select_lam(lam_grid, w1, u_tilde_hat, v_new, Z, sigma2_hat_u, nd)

        # update u
        u_tilde = get_tilde(lambda_u, u_tilde_hat, w1)  # (n, 1)
        u_new = u_tilde / np.linalg.norm(u_tilde) if np.linalg.norm(u_tilde) != 0 else u_tilde

    return u_new, v_new, lambda_u, lambda_v


def SSVD_layer(X, lam_grid, gamma1, gamma2, max_iter=5000, tol=1e-6):
    """Get the sparse SVD layer given the data matrix X at a SVD layer and the tuning parameters grid.
    Input: 
        X: (n, d), can be the original data matrix or the residual matrix
        lam_grid: ndarray, a grid of lambdas to be selected 
        gamma1, gamma2: scalar, tuning parameter
        max_iter: integer, the maximum iteration times
        tol: float, tolerance to stop the iteration
    Return: n_iter, u, v, s, lambda_u, lambda_v
        n_iter: number of iterations
        u: (n, 1), the final u at convergence 
        v: (d, 1), the final v at convergence 
        s: scalar, the singular value at convergence
        lambda_u, lambda_v: the optimal tuning parameter at convergence
    """
    # SVD
    U, _, VT = la.svd(X)
    # prepare vector Y and Z, which are fixed after given X
    Y = X.T.reshape((-1,1))  # (nd, 1)
    Z = X.reshape((-1,1))  # (nd, 1)
    
    # initial value
    u_old = U[:,0][:,None]
    v_old = VT[0][:,None]

    for i in range(max_iter):
        u_new, v_new, lambda_u, lambda_v = update_uv(u_old, v_old, X, Y, Z, gamma1, gamma2, lam_grid)
        if np.linalg.norm(u_new-u_old) < tol and np.linalg.norm(v_new-v_old) < tol:  # achieve the tolerance
            break 
        if np.all(u_new == 0) or np.all(u_new == 0):  # full shrinkage (i.e., all zeros in the vector)
            print("Warning: Full shrinkage has been achieved. Iterations stops. No further decomposition. The desired number of layers may not be achieved. ")
            break
        u_old, v_old = u_new, v_new
    n_iter = i+1  # number of iterations
    u, v = u_new, v_new  # the final u and v at convergence 
    s = (u_new.T @ X @ v_new)[0][0]
    if n_iter == max_iter:
        print("Warning: The maximum iteration has been achieved. Please consider increasing `max_iter`.")
    return n_iter, u, v, s, lambda_u, lambda_v

def SSVD(X, num_layer, lam_grid, gamma1, gamma2, max_iter=5000, tol=1e-6):
    """Get the SSVD given the data matrix X and the desired number of SSVD layers.
    Input: 
        X: (n, d), the original data matrix
        num_layer: desired number of SSVD layers
        lam_grid: ndarray, a grid of lambdas to be selected 
        gamma1, gamma2: scalar, tuning parameter
        max_iter: integer, the maximum iteration times
        tol: float, tolerance to stop the iteration
    Return: n_iters, us, vs, ss, lambda_us, lambda_vs
        n_iters: (num_layer,), number of iterations for each layer
        us: (n, num_layer), the final u at convergence for each layer
        vs: (d, num_layer), the final v at convergence for each layer
        ss: (num_layer,), the singular value at convergence for each layer
        lambda_us, lambda_vs: (num_layer,), the optimal tuning parameter at convergence for each layer
    """
    n, d = X.shape
    n_iters = np.zeros(num_layer, dtype = int)
    ss = np.zeros(num_layer)
    lambda_us = np.zeros(num_layer)
    lambda_vs = np.zeros(num_layer)
    us = np.zeros((n, num_layer))
    vs = np.zeros((d, num_layer))
    # initial value
    u = np.zeros((n, 1)); v = np.zeros((d, 1)); resi_mat = X; s = 0
    for i in range(num_layer):
        resi_mat = resi_mat - s*u@v.T
        n_iter, u, v, s, lambda_u, lambda_v = SSVD_layer(resi_mat, lam_grid, gamma1, gamma2, max_iter, tol)
        n_iters[i] = n_iter
        ss[i] = s
        lambda_us[i] = lambda_u
        lambda_vs[i] = lambda_v 
        us[:,i] = u[:,0]
        vs[:,i] = v[:,0]
        if np.all(u == 0) or np.all(v == 0):  # full shrinkage (i.e., all zeros in the vector)
            break
    return n_iters, us, vs, ss, lambda_us, lambda_vs

            


from setuptools import setup

setup(name='SSVDpkg_663proj21',
      version='0.1',
      description='SSVD algorithm for biclustering',
      author='Aihua Li, Yuxuan Chen',
      author_email='yc450@duke.edu',
      packages=['SSVDpkg_663proj21'],
      url='https://github.com/YuxuanMonta/STA663-final-project-AY',
      classifiers=[
                  'Intended Audience :: Developers',
                  'Topic :: Software Development :: Libraries :: Python Modules',
                  'License :: OSI Approved :: MIT License',
                  'Programming Language :: Python :: 3',
                  ],
      python_requires=">=3.6",
      zip_safe=False)

